#include <iostream>
#include "CifFile.h"
#include "CifParserBase.h"


static void ShowUsage(void);


int main(int argc, char** argv)
{

    string inFile;
    string sdbFile;
    string outFile;
    string diags;
    CifFile* fobjR = NULL;
    CifFileReadDef readDef;
    type readType;
    vector<string> catlist;
    bool retTest;

    CifParser* cifParserR = NULL;


    if (argc < 2)
    {
        ShowUsage();
        exit(1);
    }

    inFile = argv[1];

    outFile = inFile;
    outFile += ".cif";

    sdbFile = inFile;
    sdbFile += ".sdb";

    if (argc > 2)
    {
        if (argc == 3)
        {
            ShowUsage();
            exit(1);
        }
 
        if (strcmp(argv[2], "A") == 0)
        {
            readType = A;
        }
        else
        {
            readType = D;
        }

        for (int i = 3; i < argc; i++)
        {
            /* Add categories to the list. */
            catlist.push_back(argv[i]);
        }

        readDef.SetCategoryList(catlist, readType);
    }

//    vector<string> blocklist;
//    blocklist.Add("ADH041");

    /* Set the datablock list to accepted. */
//    readDef.SetDataBlockList(blocklist, A);

//    fobjR = new CifFile(CREATE_MODE, sdbFile, true,
//      StringCompare::eCASE_INSENSITIVE);
    fobjR = new CifFile(CREATE_MODE, sdbFile, true);

    cifParserR = new CifParser(fobjR, fobjR->GetVerbose(), readDef);

    cifParserR->Parse(inFile, diags);

    delete(cifParserR);

    if (!(diags.empty()))
    {
        cout << " Diagnostics [" << diags.size() << "] " << diags << endl;
    }

    Block& block = fobjR->GetBlock(fobjR->GetFirstBlockName());
    retTest = block.IsTablePresent("atom_site");
    cout << "Return from IsTablePresent(atom_site): " << retTest << endl;
    retTest = block.IsTablePresent("ATOM_site");
    cout << "Return from IsTablePresent(ATOM_site): " << retTest << endl;

    fobjR->Write(outFile);

    fobjR->Close();

    if (fobjR) delete fobjR;

    exit(0);

} /* End of main() */


void ShowUsage(void)
{

    cout << "Usage: CifReader input_file [readType] [category]*" << endl;

}
